#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, TensorDataset
from architecture import *
import torch.nn as nn
from torch.optim import Adam
from torch.autograd import Variable
################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):

    bs = 32
    max_length = 8192*4
    num_epochs = 50

    model_path = os.path.join(model_folder, 'model.pth')

    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    murmur_classes = ['Present', 'Unknown', 'Absent']
    num_murmur_classes = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome_classes = len(outcome_classes)


    ####################################################
    patient_files_train, patient_files_validation = train_test_split(patient_files, test_size=0.3, shuffle = True, random_state = 0)

    print("_______________________________")
    print("Preparing Train Data")
    recordings_p_train, labels_murmur_p_train, labels_outcome_p_train = load_patients(data_folder, patient_files_train)
    X_train, y1_train, y2_train = prepare_data(recordings_p_train, labels_murmur_p_train, labels_outcome_p_train, max_length)
    print("Total recordings : ",len(y1_train))
    print("# Murmur Labels : ", np.unique(y1_train, return_counts=True)[1])
    print("# Outcome Labels : ", np.unique(y2_train, return_counts=True)[1])
    print("_______________________________")

    print("_______________________________")
    print("Preparing Validation Data")
    recordings_p_validation, labels_murmur_p_validation, labels_outcome_p_validation = load_patients(data_folder, patient_files_validation)
    X_val, y1_val, y2_val = prepare_data(recordings_p_validation, labels_murmur_p_validation, labels_outcome_p_validation, max_length)
    print("Total recordings : ",len(y1_val))
    print("# Murmur Labels : ", np.unique(y1_val, return_counts=True)[1])
    print("# Outcome Labels : ", np.unique(y2_val, return_counts=True)[1])
    print("_______________________________")

    # Normalization
    mean = X_train.mean()
    std = X_train.std()

    print("Mean : ", mean)
    print("Mean : ", std)

    X_train = (X_train - mean)/std
    X_val = (X_val - mean)/std
    #______________________________________________________
    # Data Loaders
    #______________________________________________________
    train_ds = TensorDataset(X_train, y1_train, y2_train)
    train_dl = DataLoader(train_ds,batch_size=bs,shuffle=True)

    val_ds = TensorDataset(X_val, y1_val, y2_val)
    val_dl = DataLoader(val_ds,batch_size=bs,shuffle=True)

    model = ResNet18(1, ResBlock)
 
    # Define the loss function with Classification Cross-Entropy loss and an optimizer with Adam optimizer
    loss_fn = nn.CrossEntropyLoss()
    optimizer = Adam(model.parameters(), lr=0.001, weight_decay=0.0001)

    # Model Training

    train(num_epochs, train_dl, val_dl, model, optimizer, loss_fn, model_path)
    print('Finished Training')
    #####################################################

    if verbose >= 1:
        print('Done.')

# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):

    model_path = os.path.join(model_folder, 'model.pth')
    model = ResNet18(1, ResBlock)
    model.load_state_dict(torch.load(model_path))

    return model

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):

    max_length = 8192*4
    murmur_classes = ['Present', 'Unknown', 'Absent']
    outcome_classes = ['Abnormal', 'Normal']

    #murmur_classifier = model[0]
    #outcome_classifier = model[1]

    X = []

    for r_n in range(len(recordings)):

        # Single recording within patient.    
        x = recordings[r_n]

        # Pad zeros
        if len(x) < max_length:
            x = extend_ts(x, max_length)

        # Truncate   
        else:
                x = x[:max_length]

        X.append(x)

    X = np.expand_dims(X,axis=1) # [b,c,w]
    X = torch.tensor(X).float()

    X = (X - (-0.1368))/2160.49

    model.eval()
    murmur_probabilities, outcome_probabilities = model(X)

    murmur_probabilities = murmur_probabilities.cpu().detach().numpy()
    outcome_probabilities = outcome_probabilities.cpu().detach().numpy()
    #______________________________
    # Murmur Classifier
    #______________________________
    #murmur_classifier.eval()
    #murmur_probabilities = murmur_classifier(X).cpu().detach().numpy()

    pred_locs_murmur = np.argmax(murmur_probabilities, axis = 1)

    # If any murmur prediction at any location
    if 0 in pred_locs_murmur:
        idx = np.where(pred_locs_murmur == 0)[0]
        murmur_probabilities = murmur_probabilities[idx]

    murmur_probabilities = murmur_probabilities.mean(axis=0)

    # Choose label with higher probability.
    murmur_labels = np.zeros(len(murmur_classes), dtype=np.int_)
    idx = np.argmax(murmur_probabilities)
    murmur_labels[idx] = 1

    #______________________________
    # Outcome Classifier
    #______________________________
    
    #outcome_classifier.eval()
    #outcome_probabilities = outcome_classifier(X).cpu().detach().numpy()

    pred_locs_outcome = np.argmax(outcome_probabilities, axis = 1)

    # If any murmur prediction at any location
    if 0 in pred_locs_outcome:
        idx = np.where(pred_locs_outcome == 0)[0]
        outcome_probabilities = outcome_probabilities[idx]

    outcome_probabilities = outcome_probabilities.mean(axis=0)


    # Choose label with higher probability.
    outcome_labels = np.zeros(len(outcome_classes), dtype=np.int_)
    idx = np.argmax(outcome_probabilities)
    outcome_labels[idx] = 1


    # Concatenate classes, labels, and probabilities.
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((murmur_labels, outcome_labels))
    probabilities = np.concatenate((murmur_probabilities, outcome_probabilities))

    return classes, labels, probabilities

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

# Save your trained model.
def save_challenge_model(model_folder, imputer, murmur_classes, murmur_classifier, outcome_classes, outcome_classifier):
    d = {'imputer': imputer, 'murmur_classes': murmur_classes, 'murmur_classifier': murmur_classifier, 'outcome_classes': outcome_classes, 'outcome_classifier': outcome_classifier}
    filename = os.path.join(model_folder, 'model.sav')
    joblib.dump(d, filename, protocol=0)

# Extract features from the data.
def get_features(data, recordings):
    # Extract the age group and replace with the (approximate) number of months for the middle of the age group.
    age_group = get_age(data)

    if compare_strings(age_group, 'Neonate'):
        age = 0.5
    elif compare_strings(age_group, 'Infant'):
        age = 6
    elif compare_strings(age_group, 'Child'):
        age = 6 * 12
    elif compare_strings(age_group, 'Adolescent'):
        age = 15 * 12
    elif compare_strings(age_group, 'Young Adult'):
        age = 20 * 12
    else:
        age = float('nan')

    # Extract sex. Use one-hot encoding.
    sex = get_sex(data)

    sex_features = np.zeros(2, dtype=int)
    if compare_strings(sex, 'Female'):
        sex_features[0] = 1
    elif compare_strings(sex, 'Male'):
        sex_features[1] = 1

    # Extract height and weight.
    height = get_height(data)
    weight = get_weight(data)

    # Extract pregnancy status.
    is_pregnant = get_pregnancy_status(data)

    # Extract recording locations and data. Identify when a location is present, and compute the mean, variance, and skewness of
    # each recording. If there are multiple recordings for one location, then extract features from the last recording.
    locations = get_locations(data)

    recording_locations = ['AV', 'MV', 'PV', 'TV', 'PhC']
    num_recording_locations = len(recording_locations)
    recording_features = np.zeros((num_recording_locations, 4), dtype=float)
    num_locations = len(locations)
    num_recordings = len(recordings)
    if num_locations==num_recordings:
        for i in range(num_locations):
            for j in range(num_recording_locations):
                if compare_strings(locations[i], recording_locations[j]) and np.size(recordings[i])>0:
                    recording_features[j, 0] = 1
                    recording_features[j, 1] = np.mean(recordings[i])
                    recording_features[j, 2] = np.var(recordings[i])
                    recording_features[j, 3] = sp.stats.skew(recordings[i])
    recording_features = recording_features.flatten()

    features = np.hstack(([age], sex_features, [height], [weight], [is_pregnant], recording_features))

    return np.asarray(features, dtype=np.float32)


# Load patient data and labels based on given patient file paths.
def load_patients(data_folder, patient_files):

    num_patient_files = len(patient_files)

    # Murmur Class mapping.
    classes_murmur = ['Present', 'Unknown', 'Absent']
    # Outcome Class mapping
    classes_outcome = ['Abnormal', 'Normal']

    labels_murmur_p = []
    labels_outcome_p = []
    recordings_p = []

    for i in range(num_patient_files):

        # Load the current patient data and recordings.
        current_patient_data = load_patient_data(patient_files[i])
        current_recordings = load_recordings(data_folder, current_patient_data)

        # Extract murmur Label.
        label_murmur = get_murmur(current_patient_data) 
        if label_murmur in classes_murmur:
            current_label_murmur = classes_murmur.index(label_murmur)

        # Extract outcome Label.
        label_outcome = get_outcome(current_patient_data) 
        if label_outcome in classes_outcome:
            current_label_outcome = classes_outcome.index(label_outcome)

        # Per Patient Recordings and lables. 
        labels_murmur_p.append(current_label_murmur)
        labels_outcome_p.append(current_label_outcome)
        recordings_p.append(current_recordings) 

        current_patient_frequency = get_frequency(current_patient_data)

        if current_patient_frequency != 4000:
            print(current_patient_frequency)

        #auscultation_locs = get_locations(current_patient_data)
        #pat_number = patient_files[i].split('\\')[-1].split('.')[0]

    print("Total patients : ",num_patient_files)
    print("# Murmur Labels : ", np.unique(labels_murmur_p, return_counts=True)[1])
    print("# Outcome Labels : ", np.unique(labels_outcome_p, return_counts=True)[1])

    return recordings_p, labels_murmur_p, labels_outcome_p


def prepare_data(recordings, labels_murmur, labels_outcome, max_length):

    X = []
    y1 = []
    y2 = []
 
    for p_n in range(len(recordings)) :
        for r_n in range(len(recordings[p_n])):

            # Single recording within patient.    
            x = recordings[p_n][r_n]

            # Pad zeros
            if len(x) < max_length:
                x = extend_ts(x, max_length)

            # Truncate   
            else:
                 x = x[:max_length]

            X.append(x)
            y1.append(labels_murmur[p_n])
            y2.append(labels_outcome[p_n])

    X = np.expand_dims(X,axis=1) # [b,c,w]
    y1 = np.array(y1)
    y2 = np.array(y2)

    X = torch.tensor(X).float()
    y1 = torch.tensor(y1).long()
    y2 = torch.tensor(y2).long()

    return X, y1, y2

def extend_ts(ts, length):
    extended = np.zeros(length)
    siglength = np.min([length, ts.shape[0]])
    extended[:siglength] = ts[:siglength]
    return extended 


# Training function. We simply have to loop over our data iterator and feed the inputs to the network and optimize.
def train(num_epochs, train_loader, validation_loader, model, optimizer, loss_fn, model_path):
    
    best_loss = 1e9

    # Define your execution device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("The model will be running on", device, "device")
    # Convert model parameters and buffers to CPU or Cuda
    model.to(device)

    for epoch in range(num_epochs):  # loop over the dataset multiple times
        
        train_loss = 0.0

        for i, (images, labels_murmur, labels_outcome) in enumerate(train_loader, 0):
            
            # get the inputs
            images = Variable(images.to(device))
            labels_murmur = Variable(labels_murmur.to(device))
            labels_outcome = Variable(labels_outcome.to(device))

            # zero the parameter gradients
            optimizer.zero_grad()
            # predict classes using images from the training set
            outputs_murmur, outputs_outcome = model(images)
            # compute the loss based on model output and real labels
            loss_murmur = loss_fn(outputs_murmur, labels_murmur)
            loss_outcome = loss_fn(outputs_outcome, labels_outcome)

            loss = (loss_murmur + loss_outcome)/2

            # backpropagate the loss
            loss.backward()
            # adjust parameters based on the calculated gradients
            optimizer.step()

            # Let's print statistics for every 1,000 images
            train_loss += loss.item()     # extract the loss value


        # Compute and print the average accuracy fo this epoch when tested over all 10000 test images
        validation_loss = testAccuracy(model, device, validation_loader, loss_fn)
        print("_____________________________________________________")
        print('Epoch : ', epoch+1,)
        print('Training Loss : ', round(train_loss/len(train_loader),3))
        print('Validation Loss : ', round(validation_loss,3))
        print("_____________________________________________________")
        
        # we want to save the model if the accuracy is the best
        if validation_loss < best_loss:
            print("Saving model at epoch : ", epoch+1)
            saveModel(model, model_path)
            best_loss = validation_loss


# Function to save the model
def saveModel(model, model_path):
    torch.save(model.state_dict(), model_path)

# Function to test the model with the test dataset and print the accuracy for the test images
def testAccuracy(model, device, test_loader, loss_fn):
    
    model.eval()
    accuracy = 0.0
    total = 0.0
    validation_loss = 0.0
    
    with torch.no_grad():
        for data in test_loader:
            images, labels_murmur, labels_outcome = data

            images = Variable(images.to(device))
            labels_murmur = Variable(labels_murmur.to(device))
            labels_outcome = Variable(labels_outcome.to(device))

            # run the model on the test set to predict labels
            outputs_murmur, outputs_outcome = model(images)

            # compute the loss based on model output and real labels
            loss_murmur = loss_fn(outputs_murmur, labels_murmur)
            loss_outcome = loss_fn(outputs_outcome, labels_outcome)

            loss = (loss_murmur + loss_outcome)/2


            validation_loss += loss.item()     # extract the loss value
            
            # the label with the highest energy will be our prediction
            #_, predicted = torch.max(outputs.data, 1)
            #total += labels.size(0)
            #accuracy += (predicted == labels).sum().item()
    
    validation_loss = validation_loss/len(test_loader)
    # compute the accuracy over all test images
    #accuracy = accuracy / total
    return(validation_loss)